/**
 * tyrre - small "web-server" for dyndns applications
 * tyrre must be run through tcpserver from DJB's ucspi-tcp suite
 * compile with: diet -Os gcc -Os -Wall -o tyrre tyrre.c; sstrip tyrre
 * Frank Bergmann 01/2007
 */

/*
 * changes:
 * 0.8: - write http answer as one output string to prevent tcpserver
 *        splitting this response into two ip packets
 *      - return IP# even if dyndns file couldn't be written, send
 *        also string "Error" in this case
 * 0.7: - prefer output to net over output to file
 *      - use '\r\n' line termination
 * 0.6: logging output to stderr
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
 
#define BUFLEN 16384

const char str200[]="HTTP/1.0 200 OK\r\nServer: tyrre/0.8\r\nContent-Type: text/plain\r\n\r\n";
const char str302[]="HTTP/1.0 302 Found\r\nLocation: http://www.tuxad.com/\r\nContent-Type: text/html\r\n\r\n<title>302 Found</title>\r\nThe document is <a href=\"http://www.tuxad.com/\">here</a>.";
const char logdynstr[]="dyndns: ";

#include "tyrre-dyndnstable.c"
/**
 * dyndnstable sample:
#define TABLE_LEN 1
static char* map_table[TABLE_LEN][4] = {
  { "key1_xyz", "", "user1.mydomain.de", "\n120\n" }
};
 *
 */

int str_len(const char* in) {
  register const char* t=in;
  for (;;) {
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
    if (!*t) break; ++t;
  }
  return (int)(t-in);
}

int str_diffn(const char* a, const char* b, unsigned long limit) {
  register const unsigned char* s=(const unsigned char*)a;
  register const unsigned char* t=(const unsigned char*)b;
  register const unsigned char* u=t+limit;
  register int j;
  j=0;
  for (;;) { if (t>=u) break; if ((j=(*s-*t))) break; if (!*t) break; ++s; ++t; }
  return j;
}

int str_chr(const char *in, char needle) {
  register const char* t=in;
  register const char c=needle;
  for (;;) { if (!*t || *t==c) break; ++t; }
  return (int)(t-in);
}

void str_cat(char* out, const char* in) {
  register char* s=out;
  register const char* t=in;
  for (;;) { if (!*s) break; ++s; }
  for (;;) { if (!*t) break; *s++=*t++; }
  *s='\0';
}

void my_exit(const int st) {
  /* just redirect on ANY error */
  write(1, str302, str_len(str302));
  exit(st);
}

int main() {
  char ibuf[BUFLEN+1];
  char ur[BUFLEN+1];
  register char *cptr, *u;
  register char *p, *a;
  register int i;

  cptr=ibuf; u=ur;
  /**
   * we don't watch for a complete header!
   * just the first line will be parsed (and maybe only read)
   * we don't want to work with signals and timeouts (overhead here)
   */
  i=(int)read(0, cptr, BUFLEN);
  if(i<15) exit(1); /* data too short for HTTP-request, don't answer it */
  if(i==BUFLEN) {
    while((i=(int)read(0, u, BUFLEN))==BUFLEN);
  }
  /* start page will be immediately redirected */
  if(!str_diffn(cptr, "GET / HTTP/1.", 13)) my_exit(0); /* ok, quick redirect answer */
  if(str_diffn(cptr, "GET /", 5)) exit(2); /* not a HTTP-request, don't answer */
  cptr+=4;
  i=str_chr(cptr, ' '); /* find end of URL-string */
  /* check if it contains "HTTP/1." after "GET /.... " */
  if(str_diffn(cptr+i, " HTTP/1.", 8)) my_exit(4); /* error, but gracefully answer */
  cptr[i]='\0';
  u=cptr+1;
  a=getenv("TCPREMOTEIP"); /* IP# of the client */
  if(!str_diffn(u, "showip", 7)) { /* just show IP# on http://server/showip */
    write(1, str200, str_len(str200));
    if (a) write(1, a, str_len(a));
    exit(0);
  }
  /* check if the URL without leading "/" contains ":" as field separator */
  i=str_chr(u, ':');
  if(u[i]==':') {
    u[i]='\0';
    /*
     * currently we don't use the passwd and IP# fields
     *
    p=u+i+1;
    i=str_chr(p, ':');
    if(p[i]==':') {
      p[i]='\0';
      if (p[i+1]!='\0') a=p+i+1;
    }
     */
  }
  for(i=0;i<TABLE_LEN;i++) {
    if(!str_diffn(u, map_table[i][0], 9)) { /* entry found */
      cptr=map_table[i][2];
      p=map_table[i][3];
      u=ur;
      *u='\0';
      str_cat(u, str200);
      str_cat(u, a);
      write(1, u, str_len(u)); /* write HTTP-header */
      unlink(cptr);
      i=open(cptr, O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH);
      if (i>0) { /* create file for DYNDNS */
        write(2, logdynstr, str_len(logdynstr));
        write(2, cptr, str_len(cptr));
        write(2, ", ", 2);
        write(2, a, str_len(a));
        write(2, "\n", 1);
        write(i, a, str_len(a));
        write(i, p, str_len(p));
        close(i);
      } else {
        write(1, "\nError\n", 7);
      }
      exit(0);
    }
  }
  my_exit(8); /* no entry found/unknown URL: redirect and error */
  exit(0);
}
